@@ -10,24 +10,21 @@ from . import efnlp_pb2 as pb
 def split_input(encoded: List[int], num_segments: int, block_size: int) -> List[List[int]]:
 
     remainder = len(encoded) % num_segments
-    normalized_size = int((len(encoded)-remainder)/num_segments)
+    normalized_size = int((len(encoded) - remainder) / num_segments)
     if normalized_size <= block_size:
         raise ValueError("too many segments for specified block size")
 
-    segment_sizes = [
-        normalized_size + (1 if i < remainder else 0)
-        for i in range(num_segments)
-    ]
+    segment_sizes = [normalized_size + (1 if i < remainder else 0) for i in range(num_segments)]
 
     s, e = 0, segment_sizes[0]
     segments = []
     for i in range(num_segments):
-        segments.append(encoded[max(0,s-block_size):e])
+        segments.append(encoded[max(0, s - block_size) : e])
         s += segment_sizes[i]
         if i == num_segments - 1:
             e = len(encoded)
         else:
-            e += segment_sizes[i+1]
+            e += segment_sizes[i + 1]
 
     return segments
 
@@ -83,7 +80,6 @@ class Language:
 
 @dataclass
 class CharLanguage(Language):
-
     @staticmethod
     def from_corpus(C: str) -> CharLanguage:
         lang = sorted(list(set(c for c in C)))
@@ -122,7 +118,7 @@ class Sampler:
     # counts: Dict[int, int]
 
     def __init__(self, total: Optional[int] = 0, counts: Optional[Dict[int, int]] = None) -> None:
-        self.total: int = (total or 0)
+        self.total: int = total or 0
         self.counts: Dict[int, int] = {} if counts is None else {t: c for t, c in counts.items()}
 
     @staticmethod
@@ -153,7 +149,7 @@ class Sampler:
             S.counts.append(stc)
         return S
 
-    def bytes(self) -> bytes:
+    def __bytes__(self) -> bytes:
         return self.proto().SerializeToString()
 
     def __contains__(self, token: int) -> bool:
@@ -171,7 +167,7 @@ class Sampler:
     def get(self, t: int, default: int = 0) -> int:
         return self.counts.get(t, default)
 
-    def items(self) -> List[Tuple[int, int]]:
+    def items(self):  # -> List[Tuple[int, int]]:
         return self.counts.items()
 
     def __iter__(self):
@@ -209,7 +205,7 @@ class Sampler:
         raise Exception("We should not be here, sampler malformed")
 
     def probability(self, token: int) -> float:
-        return self.counts.get(token, 0)/self.total
+        return self.counts.get(token, 0) / self.total
 
 
 @dataclass
@@ -232,9 +228,9 @@ class SuffixTree:
 
     @staticmethod
     def from_proto_bytes(data: bytes) -> SuffixTree:
-            P = pb.SuffixTree()
-            P.ParseFromString(data)
-            return SuffixTree.from_proto(P)
+        P = pb.SuffixTree()
+        P.ParseFromString(data)
+        return SuffixTree.from_proto(P)
 
     @staticmethod
     def from_proto(T: pb.SuffixTree) -> SuffixTree:
@@ -259,7 +255,7 @@ class SuffixTree:
             T.prefixes.append(c.proto())
         return T
 
-    def bytes(self) -> bytes:
+    def __bytes__(self) -> bytes:
         return self.proto().SerializeToString()
 
     def __add__(self, child: SuffixTree) -> SuffixTree:
@@ -310,13 +306,17 @@ class SuffixTree:
         return self
 
     @staticmethod
-    def merge_proto(a: pb.SuffixTree, b: pb.SuffixTree) -> SuffixTree:
-        return SuffixTree.merge(
-            SuffixTree.from_proto(a), SuffixTree.from_proto(b)
+    def merge_proto_bytes(a: bytes, b: bytes) -> SuffixTree:
+        return SuffixTree.merge_proto(
+            SuffixTree.from_proto_bytes(a), SuffixTree.from_proto_bytes(b)
         )
 
     @staticmethod
-    def merge(a: SuffixTree, b: SuffixTree) -> SuffixTree:
+    def merge_proto(a: pb.SuffixTree, b: pb.SuffixTree) -> SuffixTree:
+        return SuffixTree.merge_trees(SuffixTree.from_proto(a), SuffixTree.from_proto(b))
+
+    @staticmethod
+    def merge_trees(a: SuffixTree, b: SuffixTree) -> SuffixTree:
         return a.merge(b)
 
     def merge(self, other: SuffixTree) -> SuffixTree:
@@ -399,18 +399,18 @@ class SuffixTree:
         # TODO: verify
 
 
-@dataclass
 class SuffixTreeSet:
 
-    size: int  # "vocab" size
-    depth: int = field(init=False)
-    trees: Dict[int, SuffixTree] = field(init=False)
-    sampler: Sampler = field(init=False) # for counting raw token occurrences
+    # size: int  # "vocab" size
+    # depth: int = field(init=False)
+    # trees: Dict[int, SuffixTree] = field(init=False)
+    # sampler: Sampler = field(init=False) # for counting raw token occurrences
 
-    def __post_init__(self) -> None:
-        self.depth = 0
-        self.trees = {t: SuffixTree(t) for t in range(self.size)}
-        self.sampler = Sampler()
+    def __init__(self, size: int = 0) -> None:
+        self.size: int = size
+        self.depth: int = 0
+        self.trees: Dict[int, SuffixTree] = {t: SuffixTree(t) for t in range(self.size)}
+        self.sampler: Sampler = Sampler()
 
     @staticmethod
     def from_proto_file(name: str) -> SuffixTreeSet:
@@ -419,9 +419,9 @@ class SuffixTreeSet:
 
     @staticmethod
     def from_proto_bytes(data: bytes) -> SuffixTreeSet:
-            P = pb.SuffixTreeSet()
-            P.ParseFromString(data)
-            return SuffixTreeSet.from_proto(P)
+        P = pb.SuffixTreeSet()
+        P.ParseFromString(data)
+        return SuffixTreeSet.from_proto(P)
 
     @staticmethod
     def from_proto(S: pb.SuffixTreeSet) -> SuffixTreeSet:
@@ -436,29 +436,42 @@ class SuffixTreeSet:
             S.prefixes.append(c.proto())
         return S
 
-    def bytes(self) -> bytes:
+    def __bytes__(self) -> bytes:
         return self.proto().SerializeToString()
 
-    def __contains__(self, token: int) -> SuffixTree:
+    def __contains__(self, token: int) -> bool:
         return token in self.trees
 
     def __getitem__(self, token: int) -> SuffixTree:
         return self.trees[token]
 
+    def __iter__(self):
+        return self.trees.__iter__()
+
+    def __next__(self):
+        return self.trees.__next__()
+
+    def items(self):
+        return self.trees.items()
+
     def parse(self, prefix: List[int], successor: int) -> SuffixTreeSet:
         if len(prefix) == 0:
             raise ValueError("Cannot parse empty prefix")
         self.sampler += successor
+        if prefix[-1] not in self.trees:
+            self.trees[prefix[-1]] = SuffixTree(prefix[-1])
         self.trees[prefix[-1]].parse(prefix[:-1], successor)
         self.depth = max(self.depth, self.trees[prefix[-1]].depth + 1)
         return self
 
-    def parse_all(self, tokens: List[int], block_size: int, ignore_start: bool = False) -> SuffixTreeSet:
+    def parse_all(
+        self, tokens: List[int], block_size: int, ignore_start: bool = False
+    ) -> SuffixTreeSet:
         if not ignore_start:
-            for i in range(1, block_size): # leading sub-block_size prefixes
+            for i in range(1, block_size):  # leading sub-block_size prefixes
                 self.parse(tokens[0:i], tokens[i])
-        for i in range(block_size, len(tokens)-1):
-            self.parse(tokens[i-block_size:i], tokens[i])
+        for i in range(block_size, len(tokens) - 1):
+            self.parse(tokens[i - block_size : i], tokens[i])
         return self
 
     def merge(self, other: SuffixTreeSet) -> SuffixTreeSet:
@@ -508,7 +521,9 @@ class SuffixTreeSet:
 
     def probability(self, prefix: List[int], successor: int) -> float:
         if (len(prefix) == 0) or (prefix[-1] not in self.trees):
-            return self.sampler.probability(successor)  # raw occurrence probability of the successor
+            return self.sampler.probability(
+                successor
+            )  # raw occurrence probability of the successor
         return self.trees[prefix[-1]].probability(prefix[:-1], successor)
 
     def generate(self, size: int, window: int, prompt: List[int]) -> List[int]:
